{-# LANGUAGE OverloadedStrings #-}

module Format
  ( pprHunks
  , pprHunk
  , pprDiff1
  ) where

import Types

import qualified Data.ByteString as B
import qualified Data.ByteString.Builder as BB
import qualified Data.ByteString.Internal as BI
import Data.String

pprHunkHdr :: Int -> Int -> BB.Builder
pprHunkHdr i j = fromString $ "@@ -" ++ show i ++ " +" ++ show j ++ " @@"

lineSep :: BB.Builder
lineSep = fromString "\n"

pprHunks :: [Hunk] -> BB.Builder
pprHunks = mconcat . map pprHunk

pprHunk :: Hunk -> BB.Builder
pprHunk ((i, j), toks) =
  mconcat ((pprHunkHdr i j <> lineSep) : map pprDiff1 toks)

pprDiff1 :: (Op, Tok) -> BB.Builder
pprDiff1 (op, (tok, s)) = fromString pfx <> escNewlines s <> lineSep
  where
    pfx = [opc, tc]
    opc =
      case op of
        Add -> '+'
        Keep -> ' '
        Remove -> '-'
        MineChanged -> '<'
        Original -> '='
        YourChanged -> '>'
    tc =
      if tok
        then '|'
        else '.'

escNewlines :: BS -> BB.Builder
escNewlines s
  | B.null s = mempty
  | B.head s == BI.c2w '\n' = BB.stringUtf8 "\\n" <> escNewlines (B.tail s)
  | B.head s == BI.c2w '\\' = BB.stringUtf8 "\\\\" <> escNewlines (B.tail s)
  | otherwise = BB.word8 (B.head s) <> escNewlines (B.tail s)
